import json
import logging
import os
from typing import get_type_hints

from googleapiclient.errors import HttpError

from goblet.handlers.handler import Handler
from goblet_gcp_client.client import get_default_project, get_default_location
from goblet.permissions import gcp_generic_resource_permissions


log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))

BIGQUERY_DATATYPES = {
    bool: "BOOL",
    str: "STRING",
    int: "INT64",
    float: "FLOAT64",
    list: "JSON",
    dict: "JSON",
}


class BigQueryRemoteFunction(Handler):
    """
    Cloud Big Query Remote Functions (Big Query routines) connected to cloudfunctions or
    cloudrun
    https://cloud.google.com/bigquery/docs/reference/standard-sql/remote-functions

    Allowed types for BigQuery remote functions
    https://cloud.google.com/bigquery/docs/reference/standard-sql/remote-functions#limitations

    """

    valid_backends = ["cloudfunction", "cloudfunctionv2", "cloudrun"]
    can_sync = False
    resource_type = "bqremotefunction"
    required_apis = ["bigquery", "bigqueryconnection"]
    permissions = [
        "bigquery.connections.create",
        "bigquery.connections.get",
        "bigquery.connections.delete",
        *gcp_generic_resource_permissions("bigquery", "routines"),
    ]

    def __init__(
        self,
        name,
        backend,
        versioned_clients=None,
        resources=None,
    ):
        super(BigQueryRemoteFunction, self).__init__(
            name=name,
            versioned_clients=versioned_clients,
            resources=resources,
            backend=backend,
        )
        self.connection_locations = set()

    def register(self, name, func, kwargs):
        """
        Register in handler resources
        :param name: name of resource
        :param func: function/method
        :param kwargs:
        :return:
        """
        dataset_id = kwargs["dataset_id"]
        vectorize_func = kwargs["vectorize_func"]
        max_batching_rows = kwargs["max_batching_rows"]
        kwargs = kwargs.pop("kwargs")
        location = kwargs.get("location", get_default_location())
        if location:
            self.connection_locations.add(location)
        _input, _output = self._get_hints(func, vectorize_func)
        # Routine names must contain only letters, numbers, and underscores, and be at most 256 characters long.
        routine_name = self.name + "_" + name
        routine_name = routine_name.replace("-", "_")
        self.resources[routine_name] = {
            "routine_name": routine_name,
            "dataset_id": dataset_id,
            "vectorize_func": vectorize_func,
            "max_batching_rows": max_batching_rows,
            "inputs": _input,
            "output": _output,
            "func": func,
            "location": location,
        }
        return True

    def __call__(self, request, context=None):
        """
        To be called from BigQuery routines
        :param request: <Request class>
                        Must contain X-Goblet-Name equals registered
                        name for cloudfunction as key on userDefinedContext

        :param context:
        :return: str Json repr
        """
        user_defined_context = request.json["userDefinedContext"]
        func_name = user_defined_context["X-Goblet-Name"]
        if not func_name:
            raise ValueError("No X-Goblet-Name header found")

        cloud_method = self.resources[func_name]
        if not cloud_method:
            raise ValueError(f"Method {func_name} not found")
        bq_tuples = request.json["calls"]
        if self.resources[func_name]["vectorize_func"]:
            unzipped_list = list(map(list, zip(*bq_tuples)))
            tuples_replies = cloud_method["func"](*unzipped_list)
        else:
            tuples_replies = []
            for _tuple in bq_tuples:
                tuples_replies.append(cloud_method["func"](*_tuple))
        reply = {"replies": tuples_replies}
        return json.dumps(reply)

    def _deploy(self, source=None, entrypoint=None):
        """
        Get a connection resource for Handler.name and set cloudfunction
        invoker IAM role to the service account assigned to the connection
        for all the cloudfunctions registered in this Handler
        :param source: inherited from Handler
        :param entrypoint: inherited from Handler
        :param config: inherited from Handler
        :return:
        """
        if not self.resources:
            return
        log.info("Deploying bigquery remote functions")
        for location in self.connection_locations:
            try:
                bq_query_connection = self.deploy_bigquery_connection(
                    f"{self.name}", location
                )
                self.service_accounts.append(
                    bq_query_connection["cloudResource"]["serviceAccountId"]
                )
            except HttpError as exception:
                if exception.resp.status == 409:
                    log.info(
                        "Connection already created bigquery query: for %s", self.name
                    )
                else:
                    log.error("Create connection %s", exception.error_details)
                    raise exception

        for _, resource in self.resources.items():
            create_routine_query = self.create_routine_payload(resource)
            routine_name = resource["routine_name"]
            try:
                self.versioned_clients.bigquery_routines.execute(
                    "insert",
                    params={
                        "body": create_routine_query,
                        "projectId": get_default_project(),
                        "datasetId": resource["dataset_id"],
                    },
                    parent_key="projectId",
                )
                log.info("Created bq routine %s", routine_name)
            except HttpError as exception:
                if exception.resp.status == 409:
                    self.versioned_clients.bigquery_routines.execute(
                        "update",
                        params={
                            "body": create_routine_query,
                            "projectId": get_default_project(),
                            "datasetId": resource["dataset_id"],
                            "routineId": routine_name,
                        },
                        parent_key="projectId",
                    )
                    log.info("Updated remote function %s", routine_name)
                else:
                    log.error(
                        "Bigquery remote function couldn't be created "
                        "nor updated name %s with error: %s",
                        routine_name,
                        exception.error_details,
                    )
                    raise exception

        # def _sync(self, dryrun=False):

        # ITERATE ALL DATASET TO GET ALL ROUTINES
        # :param dryrun:
        # :return:

        # client = self.versioned_clients.bigquery_routines
        # checked_dataset_id = []
        # for routine_id, routine in self.resources.items():
        #     dataset_id = routine["dataset_id"]
        #     if dataset_id in checked_dataset_id:
        #         continue
        #     checked_dataset_id.append(dataset_id)
        #     available_routines = client.execute("list",
        #     params={"datasetId": dataset_id, "projectId": get_default_project()},
        #     parent=False)
        #     if "routines" not in available_routines:
        #         continue
        #     for available_routine in available_routines["routines"]:
        #         if available_routine["routineReference"]["routineId"] not in self.resources:
        #               log.info(f'Detected unused routine in GCP
        #                {available_routine["routineReference"]["routineId"]}')
        #         if not dryrun:
        #         self.destroy_routine(dataset_id,
        #         available_routine["routineReference"]["routineId"])

    def destroy(self):
        """
        Destroy connection then destroy one by one every routine
        :return:
        """
        if not self.resources:
            return
        self.destroy_bigquery_connection()
        for _, resource in self.resources.items():
            self.destroy_routine(resource["dataset_id"], resource["routine_name"])

    def deploy_bigquery_connection(self, connection_name, location):
        """
            Creates (or get if exists) a connection resource with Handler.name
        :param connection_name: name for the connection
        :return:
        """
        connection_id = f"{self.name}"
        resource_type = {"cloudResource": {}}
        try:
            bq_connection = self.versioned_clients.bigquery_connections.execute(
                "create",
                params={"body": resource_type, "connectionId": connection_id},
                parent_schema=f"projects/{get_default_project()}/locations/{location}",
            )
            log.info(f"Created bigquery connection name: {connection_id}")

        except HttpError as exception:
            if exception.resp.status == 409:
                log.info(
                    f"Bigquery connection already exist with name: {connection_name} for {self.name} and location {location}"
                )
                client = self.versioned_clients.bigquery_connections
                bq_connection = client.execute(
                    "get",
                    params={"name": client.parent + "/connections/" + connection_id},
                    parent=False,
                )
                log.info(f"Returning connection {bq_connection['name']}")
            else:
                log.error(exception.error_details)
                raise exception
        log.info(bq_connection)
        return bq_connection

    def destroy_bigquery_connection(self):
        """
        Destroy bigquery connection, if already exist do nothing
        :return:
        """
        client = self.versioned_clients.bigquery_connections
        for location in self.connection_locations:
            try:
                client.execute(
                    "delete",
                    params={
                        "name": f"projects/{get_default_project()}/locations/{location}"
                        + "/connections/"
                        + self.name
                    },
                    parent=False,
                )
            except HttpError as exception:
                if exception.resp.status == 404:
                    log.info(f"Connection {self.name} already destroyed")
                else:
                    raise exception
            return True

    def destroy_routine(self, dataset_id, routine_id):
        """

        :param dataset_id:
        :param routine_id:
        """
        try:
            self.versioned_clients.bigquery_routines.execute(
                "delete",
                params={
                    "projectId": get_default_project(),
                    "datasetId": dataset_id,
                    "routineId": routine_id,
                },
                parent=False,
            )
            log.info(f"Destroyed routine {routine_id} for dataset {dataset_id}")
        except HttpError as exception:
            if exception.resp.status == 409:
                log.info(f"Routine {routine_id} already destroyed")
            elif exception.resp.status == 404:
                log.info(f"Routine {routine_id} doesn't exist. already destroyed?")
            else:
                log.error(
                    f"Couldn't destroy {routine_id} for dataset {dataset_id}. {exception.error_details}"
                )
                raise exception

    @staticmethod
    def _get_composite_hint(hint):
        try:
            inner = hint.__args__
        except AttributeError:
            raise AttributeError(f"Expected a composite hint, got {str(hint.__name__)}")
        return inner[0]

    def _get_hints(self, func, vectorize_func=False):
        """
        Inspect hint in function func and creates an array for input and output with SQL Datatypes
        according to https://cloud.google.com/bigquery/docs/reference/standard-sql/data-types
        restricted to
        https://cloud.google.com/bigquery/docs/reference/standard-sql/remote-functions#limitations
        :param func: function/method class to be inspected
        :return: inputs, outputs
                inputs = [{"name":<variable_name>, "data_type":<data_type>}]
                See data types on BIGQUERY_DATATYPES
        """
        type_hints = get_type_hints(func)
        inputs = []
        outputs = []

        for var_name, var_type in type_hints.items():
            if vectorize_func:
                var_type = self._get_composite_hint(var_type)
            if var_name == "return":
                outputs.append(BIGQUERY_DATATYPES[var_type])
            else:
                inputs.append(
                    {"name": var_name, "data_type": BIGQUERY_DATATYPES[var_type]}
                )
        return inputs, outputs

    def create_routine_payload(self, resource):
        """
        Create a routine object according to BigQuery specification
        :param resource: a resource saved in resources in Handler
        :param connection: a dict representing a bigquery connection
        (https://cloud.google.com/bigquery/docs/reference/bigqueryconnection/
            rest/v1/projects.locations.connections)
        :return: a dict representing a routine according to
                https://cloud.google.com/bigquery/docs/reference/rest/v2/routines
        """
        remote_function_options = {
            "endpoint": self.backend.http_endpoint,
            "connection": f"projects/{get_default_project()}/locations/{resource['location']}/connections/{self.name}",
            "userDefinedContext": {"X-Goblet-Name": resource["routine_name"]},
            "maxBatchingRows": str(resource["max_batching_rows"]),
        }
        routine_reference = {
            "projectId": get_default_project(),
            "datasetId": resource["dataset_id"],
            "routineId": resource["routine_name"],
        }

        arguments = []
        for _input in resource["inputs"]:
            argument = {
                "name": _input["name"],
                "dataType": {"typeKind": _input["data_type"]},
            }
            arguments.append(argument)
        return_type = {"typeKind": resource["output"][0]}
        language = "SQL"

        query_request = {
            "language": language,
            "routineReference": routine_reference,
            "routineType": "SCALAR_FUNCTION",
            "arguments": arguments,
            "returnType": return_type,
            "remoteFunctionOptions": remote_function_options,
        }

        return query_request



import logging
import os

from goblet.handlers.handler import Handler
from goblet_gcp_client.client import get_default_project, get_default_location

from goblet.common_cloud_actions import get_cloudrun_url
from goblet.permissions import gcp_generic_resource_permissions, add_binding

from googleapiclient.errors import HttpError

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class Scheduler(Handler):
    """Cloud Scheduler job which calls http endpoint
    https://cloud.google.com/scheduler/docs
    """

    resource_type = "scheduler"
    valid_backends = ["cloudfunction", "cloudfunctionv2", "cloudrun"]
    can_sync = True
    required_apis = ["cloudscheduler"]
    permissions = [
        *gcp_generic_resource_permissions("cloudscheduler", "jobs"),
    ]

    def register(self, name, func, kwargs):
        schedule = kwargs["schedule"]
        timezone = kwargs["timezone"]
        kwargs = kwargs.pop("kwargs")
        description = kwargs.get("description", "Created by goblet")
        headers = kwargs.get("headers", {})
        httpMethod = kwargs.get("httpMethod", "GET")
        retry_config = kwargs.get("retryConfig")
        body = kwargs.get("body")
        uri = kwargs.get("uri")
        attempt_deadline = kwargs.get("attemptDeadline")
        authMethod = kwargs.get("authMethod", "oidcToken")

        job_num = 1
        if self.resources.get(name):
            # increment job_num if there is already a scheduled job for this func
            job_num = self.resources[name]["job_num"] + 1
            self.resources[name]["job_num"] = job_num
            name = f"{name}-{job_num}"
        self.resources[name] = {
            "job_num": job_num,
            "job_json": {
                "name": f"projects/{get_default_project()}/locations/{get_default_location()}/jobs/{self.name}-{name}",
                "schedule": schedule,
                "timeZone": timezone,
                "description": description,
                "retry_config": retry_config,
                "attemptDeadline": attempt_deadline,
                "httpTarget": {
                    # "uri": ADDED AT runtime,
                    "headers": {
                        "X-Goblet-Type": "schedule",
                        "X-Goblet-Name": name,
                        **headers,
                    },
                    "body": body,
                    "httpMethod": httpMethod,
                    authMethod: {
                        # "serviceAccountEmail": ADDED AT runtime
                    },
                },
            },
            "authMethod": authMethod,
            "uri": uri,
            "func": func,
        }

    def __call__(self, request, context=None):
        headers = request.headers
        func_name = headers.get("X-Goblet-Name")
        if not func_name:
            raise ValueError("No X-Goblet-Name header found")

        job = self.resources[func_name]
        if not job:
            raise ValueError(f"Function {func_name} not found")
        return job["func"]()

    def _deploy(self, source=None, entrypoint=None):
        if not self.resources:
            return

        if self.backend.resource_type.startswith("cloudfunction"):
            resp = self.versioned_clients.cloudfunctions.execute(
                "get", parent_key="name", parent_schema=self.cloudfunction
            )
            if not resp:
                raise ValueError(f"Function {self.cloudfunction} not found")
            try:
                target = resp["httpsTrigger"]["url"]
                service_account = resp["serviceAccountEmail"]
            except KeyError:
                target = resp["serviceConfig"]["uri"]
                service_account = resp["serviceConfig"]["serviceAccountEmail"]

        if self.backend.resource_type == "cloudrun":
            # dont get target in scheduler is needed only for jobs
            cloudrun_target = None
            if self.config.cloudrun_revision and self.config.cloudrun_revision.get(
                "serviceAccount"
            ):
                service_account = self.config.cloudrun_revision.get("serviceAccount")
            elif self.config.scheduler and self.config.scheduler.get("serviceAccount"):
                service_account = self.config.scheduler.get("serviceAccount")
            elif self.config.job_spec and self.config.job_spec.get("serviceAccount"):
                service_account = self.config.job_spec.get("serviceAccount")
            else:
                raise ValueError(
                    "Service account not found in cloudrun. You can set `serviceAccount` field in config.json under `scheduler`"
                )

        self.service_accounts = [service_account]

        log.info("deploying scheduled jobs......")
        for job_name, job in self.resources.items():
            if job["uri"]:
                target = job["uri"]
            elif self.backend.resource_type.startswith("cloudfunction"):
                target = target
            else:
                # only run once
                if not cloudrun_target:
                    cloudrun_target = get_cloudrun_url(
                        self.versioned_clients.run, self.name
                    )
                target = cloudrun_target
            job["job_json"]["httpTarget"]["uri"] = target
            job["job_json"]["httpTarget"][job["authMethod"]][
                "serviceAccountEmail"
            ] = service_account

            self.deploy_job(job_name, job["job_json"])
            triggered_resource = f"{self.name}-{job_name.split('schedule-job-')[-1]}"
            # Set invoker permission on cloudrun job
            if "jobs/" in target:
                add_binding(
                    self.versioned_clients.run_job,
                    "projects/{project_id}/locations/{location_id}/jobs/"
                    + triggered_resource,
                    "roles/run.invoker",
                    [f"serviceAccount:{service_account}"],
                )

    def _sync(self, dryrun=False):
        jobs = self.versioned_clients.cloudscheduler.execute("list").get("jobs", [])
        filtered_jobs = list(
            filter(lambda job: f"jobs/{self.name}-" in job["name"], jobs)
        )
        for filtered_job in filtered_jobs:
            split_name = filtered_job["name"].split("/")[-1].split(f"{self.name}-")
            filtered_name = split_name[-1]
            if not self.resources.get(filtered_name):
                log.info(f'Detected unused job in GCP {filtered_job["name"]}')
                if not dryrun:
                    # TODO: Handle deleting multiple jobs with same name
                    self._destroy_job(filtered_name)

    def deploy_job(self, job_name, job):
        try:
            self.versioned_clients.cloudscheduler.execute(
                "create", params={"body": job}
            )
            log.info(f"created scheduled job: {job_name} for {self.name}")
        except HttpError as e:
            if e.resp.status == 409:
                log.info(f"updated scheduled job: {job_name} for {self.name}")
                self.versioned_clients.cloudscheduler.execute(
                    "patch",
                    parent_key="name",
                    parent_schema=job["name"],
                    params={"body": job},
                )
            else:
                raise e

    def destroy(self):
        if not self.resources:
            return
        for job_name in self.resources.keys():
            self._destroy_job(job_name)

    def _destroy_job(self, job_name):
        try:
            self.versioned_clients.cloudscheduler.execute(
                "delete",
                parent_key="name",
                parent_schema="projects/{project_id}/locations/{location_id}/jobs/"
                + self.name
                + "-"
                + job_name,
            )
            log.info(f"Destroying scheduled job {job_name}......")
        except HttpError as e:
            if e.resp.status == 404:
                log.info("Scheduled jobs already destroyed")
            else:
                raise e



from requests import request
import re

from goblet.backends.backend import Backend
from goblet.client import VersionedClients
from goblet_gcp_client.client import get_default_location, get_default_project
from goblet.common_cloud_actions import (
    get_function_runtime,
    create_cloudfunctionv2,
    destroy_cloudfunction,
    destroy_cloudfunction_artifacts,
    get_cloudfunction_url,
)
from goblet.errors import GobletValidationError
from goblet.permissions import gcp_generic_resource_permissions, add_binding


class CloudFunctionV2(Backend):
    """Class for cloudfunctions second generation"""

    resource_type = "cloudfunctionv2"
    supported_versions = ["v2alpha", "v2beta", "v2"]
    config_key = "cloudfunction_v2"
    monitoring_type = "cloud_function"
    monitoring_label_key = "function_name"
    required_apis = ["cloudfunctions", "cloudresourcemanager", "cloudbuild"]
    permissions = [
        "cloudfunctions.functions.getIamPolicy",
        "cloudfunctions.functions.setIamPolicy",
        "cloudfunctions.operations.get",
        "cloudfunctions.functions.sourceCodeSet",
        "cloudresourcemanager.projects.get",
        "iam.serviceaccounts.actAs",
        *gcp_generic_resource_permissions("cloudfunctions", "functions"),
    ]

    def __init__(self, app):
        self.client = VersionedClients({"cloudfunctions": "v2"}).cloudfunctions
        self.func_path = f"projects/{get_default_project()}/locations/{get_default_location()}/functions/{app.function_name}"
        super().__init__(app, self.client, self.func_path)

    def validation_config(self):
        name_pattern = r"^[a-z0-9-]+$"
        pattern = re.compile(name_pattern)
        if not re.fullmatch(pattern, self.name):
            raise GobletValidationError(
                f"Invalid Cloudfunction name {self.name}. Needs to follow regex of pattern {name_pattern}"
            )

    def deploy(self, force=False):
        put_headers = {
            "content-type": "application/zip",
        }
        source, changes = self._gcs_upload(self.client, put_headers, force=force)
        if not changes:
            return None

        if self.app.is_http():
            client, params = self._get_upload_params(source)
            create_cloudfunctionv2(client, params)

        return source

    def destroy(self, all=False):
        destroy_cloudfunction(self.client, self.name)
        if all:
            destroy_cloudfunction_artifacts(self.name)

    def _get_upload_params(self, source):
        user_configs = self.config.cloudfunction_v2 or {}
        build_configs = user_configs.get("buildConfig", {})
        if build_configs:
            del user_configs["buildConfig"]
        params = {
            "body": {
                "name": self.func_path,
                "environment": "GEN_2",
                "description": self.config.description or "created by goblet",
                "buildConfig": {
                    "runtime": get_function_runtime(self.client, self.config),
                    "entryPoint": "goblet_entrypoint",
                    "source": {"storageSource": source["storageSource"]},
                    **build_configs,
                },
                "labels": {**self.config.labels},
                **user_configs,
            },
            "functionId": self.app.function_name,
        }
        return self.client, params

    def _checksum(self):
        source_info = self.client.execute(
            "generateDownloadUrl", parent_key="name", parent_schema=self.func_path
        )
        resp = request("HEAD", source_info["downloadUrl"])
        return resp.headers["x-goog-hash"].split(",")[-1].split("=", 1)[-1]

    def update_config(self, infra_configs=[], write_config=False, stage=None):
        config_updates = {self.config_key: {}}
        for infra_config in infra_configs:
            resource_type = infra_config["resource_type"]
            if resource_type == "vpcconnector":
                config_updates[self.config_key]["serviceConfig"] = {
                    **config_updates[self.config_key].get("serviceConfig", {}),
                    "vpcConnector": infra_config["values"]["name"],
                    "vpcConnectorEgressSettings": infra_config["values"]["egress"],
                }
            elif resource_type in ("redis", "cloudtaskqueue"):
                config_updates[self.config_key]["serviceConfig"] = {
                    **config_updates[self.config_key].get("serviceConfig", {}),
                    "environmentVariables": {
                        **config_updates[self.config_key]
                        .get("serviceConfig", {})
                        .get("environmentVariables", {}),
                        **infra_config.get("values"),
                    },
                }

        self.config.update_g_config(
            values=config_updates, write_config=write_config, stage=stage
        )

    @property
    def http_endpoint(self):
        return get_cloudfunction_url(self.client, self.name)

    def set_iam_policy(self, service_account_id):
        client = self.client
        resource_name = self.func_path
        policy = {
            "policy": {
                "bindings": {
                    "role": "roles/cloudfunctions.invoker",
                    "members": [f"serviceAccount:{service_account_id}"],
                }
            }
        }
        client.execute(
            "setIamPolicy",
            params={"body": policy},
            parent_key="resource",
            parent_schema=resource_name,
        )

    def get_environment_vars(self):
        return (
            self.config.config.get("cloudfunction_v2", {})
            .get("serviceConfig", {})
            .get("environmentVariables", {})
        )

    def add_invoker_binding(self, principles):
        add_binding(
            self.client, self.func_path, "roles/cloudfunctions.invoker", principles
        )



import base64
import hashlib
import logging
import os
import warnings
import zipfile
from pathlib import Path

import requests
from googleapiclient.errors import HttpError

import goblet.globals as g
from goblet.utils import get_g_dir, checksum, build_stage_config
from goblet.common_cloud_actions import check_or_enable_service


class Backend:
    """Base backend class"""

    resource_type = ""
    version = ""
    required_files = ["main.py"]
    config_key = ""
    monitoring_type = ""
    monitoring_label_key = ""
    required_apis = []
    permissions = []

    def __init__(self, app, client, func_path):
        self.app = app
        self.name = app.function_name
        self.log = logging.getLogger("goblet.backend")
        self.log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))
        self.zip_path = get_g_dir() + f"/{self.name}.zip"
        self._zipf = None
        self.config = g.config

        # specifies which files to be zipped
        custom_files = self.config.custom_files or {}
        include = ["*.py", ".goblet/*.py", "requirements.txt"]
        exclude = ["build", "docs", "examples", "test", "tests", "venv", ".git"]

        include.extend(custom_files.get("include", []))
        exclude.extend(custom_files.get("exclude", []))

        self.zip_config = {"include": include, "exclude": exclude}

        self.func_path = func_path

        self.client = client
        self.validation_config()

    def validation_config(self):
        pass

    def deploy(self, force=False):
        raise NotImplementedError("destroy")

    def skip_deployment(self):
        return False

    def destroy(self, all=False):
        raise NotImplementedError("destroy")

    def update_config(self, infra_config={}, write_config=False, stage=None):
        raise NotImplementedError("update_config")

    @property
    def zipf(self):
        if not self._zipf:
            self._zipf = self._create_zip()
        return self._zipf

    def _create_zip(self):
        """Creates initial goblet zipfile"""
        if not os.path.isdir(get_g_dir()):
            os.mkdir(get_g_dir())
        return zipfile.ZipFile(self.zip_path, "w", zipfile.ZIP_DEFLATED)

    def delta(self, zip_path=None):
        """Compares md5 hash between local zipfile and cloudfunction already deployed"""
        if zip_path is None:
            zip_path = self.zip_path
        self.zipf.close()
        with open(zip_path, "rb") as fh:
            local_checksum = base64.b64encode(checksum(fh, hashlib.md5())).decode(
                "ascii"
            )

        deployed_checksum = self._checksum()
        modified = deployed_checksum != local_checksum
        return modified

    def _checksum(self):
        raise NotImplementedError("_checksum")

    def _gcs_upload(self, client, headers, upload_client=None, force=False):
        self.log.info("zipping source code")
        self.zip()
        if not force and self.get() and not self.delta():
            self.log.info("No changes detected....")
            return None, False
        self.log.info("uploading source zip to gs......")
        return self._upload_zip(upload_client or client, headers), True

    def _upload_zip(self, client, headers=None) -> dict:
        """Uploads zipped cloudfunction using generateUploadUrl endpoint"""
        self.zipf.close()
        with open(f".goblet/{self.name}.zip", "rb") as f:
            resp = client.execute("generateUploadUrl", params={"body": {}})
            try:
                requests.put(
                    resp["uploadUrl"],
                    data=f,
                    headers=headers,
                ).raise_for_status()
            except requests.exceptions.HTTPError as e:
                if not os.environ.get("G_HTTP_TEST") == "REPLAY":
                    raise e

        self.log.info("source code uploaded")

        return resp

    def get(self):
        """Returns backend currently deployed or None"""
        try:
            return self.client.execute(
                "get", parent_key="name", parent_schema=self.func_path
            )
        except HttpError as e:
            if e.resp.status != 404:
                raise

    @property
    def http_endpoint(self):
        raise NotImplementedError("http_endpoint")

    def zip_required_files(self):
        """Zip required files for the specified backend"""
        self._zip_config()
        if self.config.requirements_file:
            self._zip_file(self.config.requirements_file, "requirements.txt")
        else:
            self._zip_file("requirements.txt")
        if self.config.main_file:
            self._zip_file(self.config.main_file, "main.py")

    def zip(self):
        """Zips python files and any additional files based on config.custom_files"""
        self.zip_required_files()
        with warnings.catch_warnings():
            warnings.simplefilter("ignore")
            self._zip_directory()

    def _zip_file(self, filename, arcname=None):
        """skip files if not required and do not exist"""
        if not os.path.exists(filename) and filename not in self.required_files:
            return
        self.log.debug(f"Zipping file: {filename}...")
        self.zipf.write(filename, arcname)

    def _zip_config(self):
        config_path = ".goblet/config.json"
        stage_config_file = None
        if os.path.exists(config_path):
            if self.config.stage:
                stage_config_file = build_stage_config(
                    config_path=config_path, stage=self.config.stage
                )
                config_path = stage_config_file.name
            self.zipf.write(
                config_path, ".goblet/config.json", compress_type=zipfile.ZIP_DEFLATED
            )

        if stage_config_file:
            stage_config_file.close()

    def _zip_directory(self):
        exclusion_set = set(self.zip_config.get("exclude", []))
        globbed_files = []
        for pattern in self.zip_config.get("include", []):
            globbed_files.extend(Path("").rglob(pattern))
        for path in globbed_files:
            if not set(path.parts).intersection(exclusion_set):
                self.log.debug(f"Zipping file: {path}...")
                self.zipf.write(str(path))

    def get_environment_vars(self):
        raise NotImplementedError("get_environment_vars")

    def _check_or_enable_service(self, enable=False):
        return check_or_enable_service(self.required_apis, enable)

    def add_invoker_binding(self, principles):
        raise NotImplementedError("add_invoker_binding")



import logging
import os
from goblet_gcp_client import Client
from goblet_gcp_client.client import get_default_project
import goblet.globals as g

log = logging.getLogger("goblet.client")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))

DEFAULT_CLIENT_VERSIONS = {
    "cloudfunctions": "v1",
    "cloudbuild": "v1",
    "run": "v2",
    "pubsub": "v1",
    "apigateway": "v1",
    "cloudscheduler": "v1",
    "redis": "v1",
    "vpcaccess": "v1",
    "bigquery": "v2",
    "bigqueryconnection": "v1",
    "secretmanager": "v1",
    "cloudtasks": "v2",
    "serviceusage": "v1",
    "iam": "v1",
    "cloudresourcemanager": "v3",
    "artifactregistry": "v1",
    "monitoring": "v3",
    "storage": "v1",
}


def get_default_project_number():
    client = Client("cloudresourcemanager", "v1", calls="projects")
    resp = client.execute(
        "get", parent_key="projectId", parent_schema=get_default_project()
    )
    return resp["projectNumber"]


# Clients
class VersionedClients:
    def __init__(self, client_versions=None):
        self.client_versions = DEFAULT_CLIENT_VERSIONS
        if (
            g.config
            and g.config.client_versions
            and isinstance(g.config.client_versions, dict)
        ):
            self.client_versions.update(g.config.client_versions)
        if client_versions:
            self.client_versions.update(client_versions)

    @property
    def cloudfunctions(self):
        return Client(
            "cloudfunctions",
            self.client_versions.get("cloudfunctions", "v1"),
            calls="projects.locations.functions",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def cloudbuild(self):
        return Client(
            "cloudbuild",
            self.client_versions.get("cloudbuild", "v1"),
            calls="projects.builds",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def run(self):
        return Client(
            "run",
            self.client_versions.get("run", "v2"),
            calls="projects.locations.services",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def run_job(self):
        return Client(
            "run",
            self.client_versions.get("run", "v2"),
            calls="projects.locations.jobs",
            parent_schema="projects/{project_id}/locations/{location_id}",
            regional=True,
        )

    @property
    def pubsub(self):
        return Client(
            "pubsub",
            self.client_versions.get("pubsub", "v1"),
            calls="projects.subscriptions",
            parent_schema="projects/{project_id}",
            emulator_host=os.environ.get("PUBSUB_EMULATOR_HOST"),
        )

    @property
    def pubsub_topic(self):
        return Client(
            "pubsub",
            self.client_versions.get("pubsub", "v1"),
            calls="projects.topics",
            parent_schema="projects/{project_id}",
            emulator_host=os.environ.get("PUBSUB_EMULATOR_HOST"),
        )

    @property
    def apigateway(self):
        return Client(
            "apigateway",
            self.client_versions.get("apigateway", "v1"),
            calls="projects.locations.gateways",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def apigateway_configs(self):
        return Client(
            "apigateway",
            self.client_versions.get("apigateway", "v1"),
            calls="projects.locations.apis.configs",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def apigateway_api(self):
        return Client(
            "apigateway",
            self.client_versions.get("apigateway", "v1"),
            calls="projects.locations.apis",
            parent_schema="projects/{project_id}/locations/global",
        )

    @property
    def cloudscheduler(self):
        return Client(
            "cloudscheduler",
            self.client_versions.get("cloudscheduler", "v1"),
            calls="projects.locations.jobs",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def cloudtask(self):
        return Client(
            "cloudtasks",
            self.client_versions.get("cloudtasks", "v2"),
            calls="projects.locations.queues.tasks",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def cloudtask_queue(self):
        return Client(
            "cloudtasks",
            self.client_versions.get("cloudtasks", "v2"),
            calls="projects.locations.queues",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def eventarc(self):
        return Client(
            "eventarc",
            self.client_versions.get("eventarc", "v1"),
            calls="projects.locations.triggers",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def redis(self):
        return Client(
            "redis",
            self.client_versions.get("redis", "v1"),
            calls="projects.locations.instances",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def vpcconnector(self):
        return Client(
            "vpcaccess",
            self.client_versions.get("vpcaccess", "v1"),
            calls="projects.locations.connectors",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def bigquery_connections(self):
        return Client(
            "bigqueryconnection",
            self.client_versions.get("bigqueryconnection", "v1"),
            calls="projects.locations.connections",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def run_uploader(self):
        return Client(
            "cloudfunctions",
            "v2beta",
            calls="projects.locations.functions",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def bigquery_routines(self):
        return Client(
            "bigquery",
            self.client_versions.get("bigquery", "v2"),
            calls="routines",
            parent_schema="{project_id}",
        )

    @property
    def monitoring_alert(self):
        return Client(
            "monitoring",
            self.client_versions.get("monitoring", "v3"),
            calls="projects.alertPolicies",
            parent_schema="projects/{project_id}",
        )

    @property
    def monitoring_uptime(self):
        return Client(
            "monitoring",
            self.client_versions.get("monitoring", "v3"),
            calls="projects.uptimeCheckConfigs",
            parent_schema="projects/{project_id}",
        )

    @property
    def logging_metric(self):
        return Client(
            "logging",
            self.client_versions.get("logging", "v2"),
            calls="projects.metrics",
            parent_schema="projects/{project_id}",
        )

    @property
    def secretmanager(self):
        return Client(
            "secretmanager",
            self.client_versions.get("secretmanager", "v1"),
            calls="projects.secrets.versions",
            parent_schema="projects/{project_id}",
        )

    @property
    def service_usage(self):
        return Client(
            "serviceusage",
            self.client_versions.get("serviceusage", "v1"),
            calls="services",
            parent_schema="projects/{project_id}",
        )

    @property
    def iam_roles(self):
        return Client(
            "iam",
            self.client_versions.get("iam", "v1"),
            calls="projects.roles",
            parent_schema="projects/{project_id}",
        )

    @property
    def service_account(self):
        return Client(
            "iam",
            self.client_versions.get("iam", "v1"),
            calls="projects.serviceAccounts",
            parent_schema="projects/{project_id}",
        )

    @property
    def project_resource_manager(self):
        return Client(
            "cloudresourcemanager",
            self.client_versions.get("cloudresourcemanager", "v3"),
            calls="projects",
            parent_schema="projects/{project_id}",
        )

    @property
    def artifactregistry_repositories(self):
        return Client(
            "artifactregistry",
            self.client_versions.get("artifactregistry", "v1"),
            calls="projects.locations.repositories",
            parent_schema="projects/{project_id}/locations/{location_id}",
        )

    @property
    def storage_buckets(self):
        return Client(
            "storage",
            self.client_versions.get("storage", "v1"),
            calls="buckets",
        )

    @property
    def storage_objects(self):
        return Client(
            "storage",
            self.client_versions.get("storage", "v1"),
            calls="objects",
        )



import json
import logging
import sys
import os

from goblet.config import GConfig
import goblet.globals as g
from goblet.decorators import Goblet_Decorators
from goblet.resource_manager import Resource_Manager

from google.cloud.logging.handlers import StructuredLogHandler
from google.cloud.logging_v2.handlers import setup_logging
from typing import List

logging.basicConfig()

log = logging.getLogger("goblet.app")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class Goblet(Goblet_Decorators, Resource_Manager):
    """
    Main class which inherits most of its logic from the Register_Handlers class. Local param is used
    to set the entrypoint for running goblet locally
    """

    def __init__(
        self,
        function_name="goblet",
        backend="cloudfunction",
        local="local",
        cors=None,
        routes_type="apigateway",
        config=None,
        log_level="INFO",
        labels={},
        is_sub_app=False,
    ):
        g.config = GConfig(config or {})
        self.config = g.config
        self.function_name = self.config.function_name or function_name
        self.labels = labels

        self.backend_class = self.get_backend_and_check_versions(backend)
        self.backend = self.backend_class(self)
        self.is_sub_app = is_sub_app

        super(Goblet, self).__init__(
            function_name=self.function_name,
            backend=self.backend,
            cors=cors,
            routes_type=routes_type,
        )
        self.log = logging.getLogger(__name__)
        self.headers = {}
        self.g = G()

        # Setup Local
        module_name = GConfig(config).main_file or "main"
        module_name = module_name.replace(".py", "")
        if local and sys.modules.get(module_name):

            def local_func(request):
                return self(request)

            setattr(sys.modules[module_name], local, local_func)

        # configure logging for local or gcp
        if os.environ.get("X_GOBLET_LOCAL") or os.environ.get("G_HTTP_TEST"):
            logging.basicConfig()
            self.log = logging.getLogger("werkzeug")
        elif not os.environ.get("X-GOBLET-DEPLOY"):
            self.log.handlers.clear()
            handler = StructuredLogHandler()
            setup_logging(
                handler,
                log_level=logging.getLevelName(
                    os.getenv("GOBLET_LOG_LEVEL", log_level)
                ),
            )
            self.log = logging.getLogger(__name__)

    def deploy(
        self,
        skip_handlers=False,
        skip_backend=False,
        skip_infra=False,
        force=False,
        write_config=False,
        stage=None,
        handlers=None,
        infras=None,
    ):
        g.config.update_g_config(values={"labels": self.labels})
        source = None
        backend = self.backend

        if infras or not skip_infra:
            log.info("deploying infrastructure")
            self.deploy_infrastructure(infras)

        infra_config = self.get_infrastructure_config()
        backend.update_config(infra_config, write_config, stage)

        if not skip_backend:
            log.info(f"preparing to deploy with backend {self.backend.resource_type}")
            source = backend.deploy(force=force)

        registered_handlers = self.get_registered_handler_resource_types()

        # checks registered deployable handlers before determining if backend exists
        if (
            registered_handlers
            and skip_backend
            and (handlers or not skip_handlers)
            and not backend.skip_deployment()
            and not backend.get()
        ):
            log.error("backend is not deployed, handlers cannot be deployed. exiting.")
            sys.exit(1)

        if handlers or not skip_handlers:
            log.info("deploying handlers")
            self.deploy_handlers(source, handlers)

    def destroy(
        self,
        all=False,
        skip_infra=False,
        skip_handlers=False,
        skip_backend=False,
        handlers: List[str] = None,
        infras: List[str] = None,
    ):
        """Destroys http cloudfunction and then calls goblet.destroy() to remove handler's infrastructure"""

        if handlers or not skip_handlers:
            log.info("destroying handlers")
            self.destroy_handlers(handlers)

        if not skip_backend:
            self.backend.destroy(all=all)

        if infras or not skip_infra:
            log.info("destroying infrastructure")
            self.destroy_infrastructure(infras)

    def sync(
        self,
        dryrun=False,
        skip_infra=False,
        skip_handlers=False,
        handlers: List[str] = None,
        infras: List[str] = None,
    ):
        if infras or not skip_infra:
            log.info("syncing infrastructure")
            self.sync_infrastructure(dryrun, infras)
        if handlers or not skip_handlers:
            log.info("syncing handlers")
            self.sync_handlers(dryrun, handlers)

    def check_or_enable_services(self, enable=False):
        self.backend._check_or_enable_service(enable)
        for _, v in self.handlers.items():
            v._check_or_enable_service(enable)
        for _, v in self.infrastructure.items():
            v._check_or_enable_service(enable)
        return None

    def get_permissions(self):
        permissions = set()
        permissions.update(self.backend.permissions)
        for _, v in self.handlers.items():
            permissions.update(v.get_permissions())
        for _, v in self.infrastructure.items():
            permissions.update(v.get_permissions())

        permissions = list(permissions)
        permissions.sort()
        return permissions

    def package(self):
        self.backend.zip()

    def deploy_local(self):
        source = None

        log.info("deploying infrastructure locally...")
        self.deploy_infrastructure()

        self.backend.skip_deployment()
        log.info("deploying handlers locally...")
        self.deploy_handlers(source)


def jsonify(*args, **kwargs):
    """
    Helper based on flask jsonify and helps convert lists and dicts into valid reponses.
    """
    indent = None
    separators = (",", ":")
    headers = {"Content-Type": "application/json"}
    headers.update(kwargs.pop("headers", {}))

    options = kwargs.pop("options", {})

    if args and kwargs:
        raise TypeError("jsonify() behavior undefined when passed both args and kwargs")
    elif len(args) == 1:  # single args are passed directly to dumps()
        data = args[0]
    else:
        data = args or kwargs

    if not isinstance(data, (str, bytes)):
        data = json.dumps(data, indent=indent, separators=separators, **options)
    return (data, 200, headers)


class G:
    """
    Global class that allows users to set and pass variables between middlewares.
    """

    pass


def goblet_entrypoint(app, entrypoint="goblet_entrypoint"):
    """
    Wrapper around Goblet app to make it a function, since a recent GCP bug forces the entrypoint to be of type function
    whereas before any type of callable was allowed.
    """
    if sys.modules.get("main"):

        def goblet_entrypoint_wrapper(request, context=None):
            return app(request, context)

        setattr(sys.modules["main"], entrypoint, goblet_entrypoint_wrapper)



from goblet.handlers.handler import Handler


class HTTP(Handler):
    """Http Trigger"""

    resource_type = "http"
    valid_backends = ["cloudfunction", "cloudfunctionv2", "cloudrun"]

    def __init__(
        self, name, backend, versioned_clients=None, cors=None, resources=None
    ):
        super(HTTP, self).__init__(
            name=name,
            versioned_clients=versioned_clients,
            resources=resources,
            backend=backend,
        )
        self.resources = resources or []

    def register(self, name, func, kwargs):
        self.resources.append({"func": func, "headers": kwargs.get("headers", {})})

    def __call__(self, request, context=None):
        headers = request.headers or {}
        for http_endpoint in self.resources:
            endpoint_headers = http_endpoint["headers"]
            if (
                isinstance(endpoint_headers, dict)
                and endpoint_headers.items() <= dict(headers.items()).items()
            ):
                return http_endpoint["func"](request)
            if isinstance(endpoint_headers, set) and endpoint_headers <= set(
                headers.keys()
            ):
                return http_endpoint["func"](request)

    def _deploy(self, source=None, entrypoint=None):
        return

    def destroy(self):
        return



from typing import Any, Optional

from apispec import BasePlugin, APISpec
from pydantic import BaseModel
from pydantic._internal._model_construction import ModelMetaclass
import copy


class PydanticPlugin(BasePlugin):
    def __init__(self) -> None:
        super().__init__()
        self.spec: APISpec | None = None

    def init_spec(self, spec: APISpec) -> None:
        super().init_spec(spec)
        self.spec = spec

    def schema_helper(
        self,
        name: str,
        definition: dict,
        model: Optional[BaseModel] = None,
        **kwargs: Any,
    ):
        if model is None:
            return None

        org_schema = model.model_json_schema(ref_template="#/definitions/{model}")
        schema = copy.deepcopy(org_schema)
        if "$defs" in schema:
            for k, v in schema["$defs"].items():
                if k not in self.spec.components.schemas:
                    self.spec.components.schema(k, self.resolve_schema(v))
            del schema["$defs"]

        return self.resolve_schema(schema)

    def operation_helper(self, path=None, operations=None, **kwargs) -> None:
        self.resolve_operation_parameters(operations)
        return

    def resolve_operation_parameters(self, operations):
        """
        Handle pydantic paramters in operations
        """
        for operation in operations.values():
            if not isinstance(operation, dict):
                continue
            if "parameters" in operation:
                operation["parameters"] = self.resolve_parameters(
                    operation["parameters"]
                )

    def resolve_schema(self, schema):
        if isinstance(schema, dict):
            return self._resolve_schema_values(schema)
        return schema

    def _resolve_schema_values(self, schema: dict):
        """Extract properties from pydantic schema"""
        if schema.get("type") == "object" and "properties" in schema:
            schema["properties"] = {
                k: self._resolve_schema_values(v)
                for k, v in schema.get("properties", {}).items()
            }
        if schema.get("type") == "array" and "items" in schema:
            schema["items"] = self._resolve_schema_values(schema["items"])

        for op in ["anyOf", "oneOf", "allOf"]:
            if op in schema:
                for k, v in schema.get(op)[0].items():
                    schema[k] = v
                del schema[op]
        return schema

    def resolve_parameters(self, parameters):
        """
        Handle Pydantic class parameters
        """
        resolved = []
        for parameter in parameters:
            if (
                isinstance(parameter, dict)
                and issubclass(type(parameter.get("schema", {})), ModelMetaclass)
                and "in" in parameter
            ):
                schema_instance = parameter.pop("schema")
                model_params = self.resolve_pydantic_model(schema_instance)
                resolved.extend(model_params)
            else:
                resolved.append(parameter)
        return resolved

    def resolve_pydantic_model(self, model):
        """
        Extract Properties from pydantic model and convert into query params
        """
        schema = model.model_json_schema()
        resolved_schema = self.resolve_schema(schema)
        params = []
        for key, value in resolved_schema["properties"].items():
            param = {
                "in": "query",
                "name": key,
                "type": value["type"],
                "required": key in schema.get("required", []),
            }
            if value["type"] == "array":
                param["items"] = value["items"]
            params.append(param)
        return params



from __future__ import annotations
import os
import yaml
from warnings import warn
import logging

from goblet_gcp_client.client import get_default_location, get_default_project

from goblet.backends.cloudfunctionv1 import CloudFunctionV1
from goblet.backends.cloudfunctionv2 import CloudFunctionV2
from goblet.backends.cloudrun import CloudRun

from goblet.infrastructures.redis import Redis
from goblet.infrastructures.vpcconnector import VPCConnector
from goblet.infrastructures.cloudtask import CloudTaskQueue
from goblet.infrastructures.pubsub import PubSubTopic
from goblet.infrastructures.alerts import PubSubDLQCondition
from goblet.infrastructures.bq_spark_stored_procedure import (
    BigQuerySparkStoredProcedure,
)

log = logging.getLogger(__name__)
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))

EVENT_TYPES = [
    "all",
    "http",
    "schedule",
    "pubsub",
    "storage",
    "route",
    "eventarc",
    "job",
    "bqremotefunction",
    "cloudtasktarget",
]

SUPPORTED_BACKENDS = {
    "cloudfunction": CloudFunctionV1,
    "cloudfunctionv2": CloudFunctionV2,
    "cloudrun": CloudRun,
}

SUPPORTED_INFRASTRUCTURES = {
    "redis": Redis,
    "vpcconnector": VPCConnector,
    "cloudtaskqueue": CloudTaskQueue,
    "pubsub_topic": PubSubTopic,
    "bqsparkstoredprocedure": BigQuerySparkStoredProcedure,
}


class Goblet_Decorators:
    """Decorator endpoints that are called by the user. Returns _create_registration_function which will trigger the corresponding
    registration function in the Register_Handlers class. For example _create_registration_function with type route will call
    _register_route"""

    def before_request(self, event_type="all"):
        """Function called before request for preeprocessing"""

        if event_type not in EVENT_TYPES:
            raise ValueError(f"{event_type} not in {EVENT_TYPES}")

        def _middleware_wrapper(func):
            self.register_middleware(func, event_type, before_or_after="before")
            return func

        return _middleware_wrapper

    def after_request(self, event_type="all"):
        """Function called after request for postprocessing"""

        if event_type not in EVENT_TYPES:
            raise ValueError(f"{event_type} not in {EVENT_TYPES}")

        def _middleware_wrapper(func):
            self.register_middleware(func, event_type, before_or_after="after")
            return func

        return _middleware_wrapper

    def middleware(self, event_type="all"):
        """Middleware functions that are called before events for preprocessing.
        This is deprecated and will be removed in the future. Use before_request instead
        """
        warn(
            "Middleware method is deprecated. Use before_request instead",
            DeprecationWarning,
            stacklevel=2,
        )

        if event_type not in EVENT_TYPES:
            raise ValueError(f"{event_type} not in {EVENT_TYPES}")

        def _middleware_wrapper(func):
            self.register_middleware(func, event_type)
            return func

        return _middleware_wrapper

    def route(self, path, methods=["GET"], **kwargs):
        """Api Gateway route"""
        return self._create_registration_function(
            handler_type="route",
            registration_kwargs={"path": path, "methods": methods, "kwargs": kwargs},
        )

    def schedule(self, schedule, timezone="UTC", **kwargs):
        """Scheduler job Http trigger"""
        return self._create_registration_function(
            handler_type="schedule",
            registration_kwargs={
                "schedule": schedule,
                "timezone": timezone,
                "kwargs": kwargs,
            },
        )

    def bqremotefunction(
        self, dataset_id, vectorize_func=False, max_batching_rows=0, **kwargs
    ):
        """
        BigQuery remote function trigger
        dataset_id: Where the function will be registered
        vectorize_func: If True, ensure every argument of your function is a list, and returns a list
        max_batching_rows: Max number of rows in each batch sent to the remote service. 0 for dynamic
        """
        return self._create_registration_function(
            handler_type="bqremotefunction",
            registration_kwargs={
                "dataset_id": dataset_id,
                "vectorize_func": vectorize_func,
                "max_batching_rows": max_batching_rows,
                "kwargs": kwargs,
            },
        )

    def pubsub_subscription(self, topic, **kwargs):
        """Pubsub topic trigger"""
        dlq = kwargs.pop("dlq", False)
        dlq_topic_config = kwargs.pop("dlq_topic_config", {})
        dlq_alert = kwargs.pop("dlq_alert", False)
        dlq_alert_config = kwargs.pop("dlq_alert_config", {})
        if dlq:
            log.info(f"DLQ enabled use of subscription will be forced to topic {topic}")
            kwargs["use_subscription"] = True
            dlq_topic_name = (
                f"{topic}-dlq"
                if "name" not in dlq_topic_config
                else dlq_topic_config.pop("name")
            )
            dlq_pull_subscription_config = dlq_topic_config.pop(
                "pull_subscription_config", {}
            )
            dlq_pull_subscription_name = (
                f"{dlq_topic_name}-pull-subscription"
                if "name" not in dlq_pull_subscription_config
                else dlq_pull_subscription_config.pop("name")
            )
            # Create DLQ topic
            self._register_infrastructure(
                handler_type="pubsub_topic",
                kwargs={
                    "name": dlq_topic_name,
                    "kwargs": {
                        "config": dlq_topic_config,
                        "dlq": True,
                        "dlq_pull_subscription": {
                            "name": dlq_pull_subscription_name,
                            "config": dlq_pull_subscription_config,
                        },
                    },
                },
            )
            dlq_policy = {
                "deadLetterPolicy": {
                    "deadLetterTopic": self.infrastructure["pubsub_topic"].resources[
                        dlq_topic_name
                    ]["name"],
                }
            }
            if "config" in kwargs:
                kwargs["config"].update(dlq_policy)
            else:
                kwargs["config"] = dlq_policy

            if dlq_alert:
                sub_name = f"{self.function_name}-{topic}"
                log.info(f"DLQ alert enabled for subscription {sub_name}")
                dlq_alert_name = (
                    f"{sub_name}-dlq-alert"
                    if "name" not in dlq_alert_config
                    else dlq_alert_config.pop("name")
                )
                dlq_alert_trigger_value = (
                    0
                    if "trigger_value" not in dlq_alert_config
                    else dlq_alert_config.pop("trigger_value")
                )
                dlq_alert_config["conditions"] = [
                    PubSubDLQCondition(
                        "pubsub",
                        subscription_id=sub_name,
                        value=dlq_alert_trigger_value,
                    )
                ]
                self._register_infrastructure(
                    handler_type="alerts",
                    kwargs={"name": dlq_alert_name, "kwargs": dlq_alert_config},
                )

        return self._create_registration_function(
            handler_type="pubsub",
            registration_kwargs={"topic": topic, "kwargs": kwargs},
        )

    def topic(self, topic, **kwargs):
        warn(
            "This method is deprecated, use @app.pubsub_subscription",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.pubsub_subscription(topic, **kwargs)

    def cloudtasktarget(self, name, **kwargs):
        """CloudTask trigger"""
        kwargs["name"] = name
        return self._create_registration_function(
            handler_type="cloudtasktarget",
            registration_kwargs={"name": name, "kwargs": kwargs},
        )

    def storage(self, bucket, event_type, name=None):
        """Storage event trigger"""
        return self._create_registration_function(
            handler_type="storage",
            registration_kwargs={
                "bucket": bucket,
                "event_type": event_type,
                "name": name,
            },
        )

    def eventarc(self, topic=None, event_filters=[], **kwargs):
        """Eventarc trigger"""
        return self._create_registration_function(
            handler_type="eventarc",
            registration_kwargs={
                "topic": topic,
                "event_filters": event_filters,
                "kwargs": kwargs,
            },
        )

    def uptime(self, **kwargs):
        """Uptime trigger"""
        return self._create_registration_function(
            handler_type="uptime",
            registration_kwargs={
                "kwargs": kwargs,
            },
        )

    def http(self, headers={}):
        """Base http trigger"""
        return self._create_registration_function(
            handler_type="http",
            registration_kwargs={"headers": headers},
        )

    def job(self, name, task_id=0, schedule=None, timezone="UTC", **kwargs):
        """Cloudrun Job"""
        if schedule and task_id != 0:
            raise ValueError("Schedule can only be added to task_id with value 0")
        if kwargs and task_id != 0:
            raise ValueError("Arguments can only be added to task_id with value 0")
        if schedule:
            self._register_handler(
                "schedule",
                f"schedule-job-{name}",
                None,
                kwargs={
                    "schedule": schedule,
                    "timezone": timezone,
                    "kwargs": {
                        "uri": f"https://{get_default_location()}-run.googleapis.com/apis/run.googleapis.com/v1/namespaces/{get_default_project()}/jobs/{self.function_name}-{name}:run",
                        "httpMethod": "POST",
                        "authMethod": "oauthToken",
                        **kwargs,
                    },
                },
            )
        return self._create_registration_function(
            handler_type="jobs",
            registration_kwargs={"name": name, "task_id": task_id, "kwargs": kwargs},
        )

    def apigateway(self, name, backend_url, filename=None, openapi_dict=None, **kwargs):
        """Api Gateway Infrastructure with an existing openapi spec.
        Requires either a filename or openapi_dict"""
        if not filename and not openapi_dict:
            raise ValueError(
                "Either a filename or the openapi_dict needs to be provided"
            )
        if filename and openapi_dict:
            raise ValueError(
                "Only one of either a filename or the openapi_dict needs to be provided"
            )
        if filename:
            with open(filename) as f:
                openapi_dict = yaml.safe_load(f.read())
        return self._register_infrastructure(
            handler_type="apigateway",
            kwargs={
                "name": name,
                "kwargs": {
                    "backend_url": backend_url,
                    "openapi_dict": openapi_dict,
                    **kwargs,
                },
            },
        )

    def alert(self, name, conditions, **kwargs):
        """Alert Resource"""
        kwargs["conditions"] = conditions
        return self._register_infrastructure(
            handler_type="alerts",
            kwargs={"name": name, "kwargs": kwargs},
        )

    def cloudtaskqueue(self, name, config=None, **kwargs):
        """CloudTask Queue Infrastructure"""
        kwargs["config"] = config
        return self._register_infrastructure(
            handler_type="cloudtaskqueue",
            kwargs={"name": name, "config": config, "kwargs": kwargs},
        )

    def pubsub_topic(self, name, config=None, **kwargs):
        """PubSub Topic Infrastructure"""
        kwargs["config"] = config
        return self._register_infrastructure(
            handler_type="pubsub_topic",
            kwargs={"name": name, "config": config, "kwargs": kwargs},
        )

    def redis(self, name, **kwargs):
        """Redis Infrastructure"""
        return self._register_infrastructure(
            handler_type="redis",
            kwargs={"name": name, "kwargs": kwargs},
        )

    def vpcconnector(self, name, **kwargs):
        """VPC Connector Infrastructure"""
        return self._register_infrastructure(
            handler_type="vpcconnector",
            kwargs={"name": name, "kwargs": kwargs},
        )

    def bqsparkstoredprocedure(
        self,
        name,
        dataset_id,
        runtime_version="1.1",
        container_image=None,
        func=None,
        spark_file=None,
        additional_python_files=None,
        additional_files=None,
        properties=None,
        **kwargs,
    ):
        """
        BigQuery Spark Stored Procedure trigger
        :param name: name of resource
        :param dataset_id: dataset id where the routine will be created
        :param func (optional): function/method
        :param runtime_version (optional): runtime version of the spark code
        :param container_image (optional): container image to use
        :param spark_file (optional): file from a local path with the spark code
        :param additional_python_files (optional): List of files from a local path with additional code (Ex: libraries)
        :param additional_files (optional): List of files from a local path with additional files (Ex: csvs)
        :param properties (optional): Dictionary with additional properties. Supported properties: https://spark.apache.org/docs/latest/configuration.html#spark-properties
        """
        return self._register_infrastructure(
            handler_type="bqsparkstoredprocedure",
            kwargs={
                "name": name,
                "kwargs": {
                    "dataset_id": dataset_id,
                    "func": func,
                    "runtime_version": runtime_version,
                    "container_image": container_image,
                    "spark_file": spark_file,
                    "additional_python_files": additional_python_files,
                    "additional_files": additional_files,
                    "properties": properties,
                    **kwargs,
                },
            },
        )

    def errorhandler(self, error):
        def _register_error_handler(error_handler):
            self.error_handlers[error] = error_handler
            return error_handler

        return _register_error_handler

    def stage(self, stage=None, stages=[]):
        if not stage and not stages:
            raise ValueError("One of stage or stages should be set")

        # Only registers if stage matches.
        def _register_stage(func):
            if os.getenv("STAGE") == stage or os.getenv("STAGE") in stages:
                return func

        return _register_stage

    def _create_registration_function(self, handler_type, registration_kwargs=None):
        def _register_handler(user_handler):
            if user_handler:
                handler_name = user_handler.__name__
                kwargs = registration_kwargs or {}
                self._register_handler(handler_type, handler_name, user_handler, kwargs)
            return user_handler

        return _register_handler

    def _register_handler(self, handler_type, name, func, kwargs, options=None):
        name = kwargs.get("kwargs", {}).get("name") or name
        self.handlers[handler_type].register(name=name, func=func, kwargs=kwargs)

    def _register_infrastructure(self, handler_type, kwargs, options=None):
        return self.infrastructure[handler_type].register(
            kwargs["name"], kwargs=kwargs.get("kwargs", {})
        )

    def register_middleware(self, func, event_type="all", before_or_after="before"):
        middleware_list = self.middleware_handlers[before_or_after].get(event_type, [])
        middleware_list.append(func)
        self.middleware_handlers[before_or_after][event_type] = middleware_list



from goblet.common_cloud_actions import (
    create_eventarc_trigger,
    destroy_eventarc_trigger,
)
from goblet.response import Response
import logging
import os

from goblet.handlers.handler import Handler
from goblet_gcp_client.client import get_default_project, get_default_location
from goblet.permissions import gcp_generic_resource_permissions

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class EventArc(Handler):
    """Eventarc trigger
    https://cloud.google.com/eventarc/docs
    """

    resource_type = "eventarc"
    # Cloudfunctions gen 2 is also supported (see https://cloud.google.com/functions/docs/calling/eventarc)
    # However, the implementation is complicated because it is difficult to route the responses to the correct trigger
    # Partial implementation can be found here: https://github.com/samdevo/goblet/blob/eventarc-changes/goblet/resources/eventarc.py
    valid_backends = ["cloudrun"]
    can_sync = True
    required_apis = ["eventarc"]
    permissions = [*gcp_generic_resource_permissions("eventarc", "triggers")]

    def __init__(self, name, backend, versioned_clients=None, resources=None):
        super(EventArc, self).__init__(
            name,
            versioned_clients=versioned_clients,
            resources=resources,
            backend=backend,
        )
        self.resources = resources or []

    def register(self, name, func, kwargs):
        event_filters = kwargs.get("event_filters")
        region = kwargs.get("kwargs", {}).get("region", get_default_location())
        event_data_content_type = kwargs.get("kwargs", {}).get(
            "event_data_content_type", "application/json"
        )
        if kwargs.get("topic"):
            event_filters = [
                {
                    "attribute": "type",
                    "value": "google.cloud.pubsub.topic.v1.messagePublished",
                }
            ]
        self.resources.append(
            {
                "trigger_name": f"{self.name}-{name}".replace("_", "-"),
                "event_filters": event_filters,
                "topic": kwargs["topic"],
                "region": region,
                "name": name,
                "func": func,
                "event_data_content_type": event_data_content_type,
            }
        )

    def __call__(self, request):
        full_path = request.path
        trigger_name = full_path.split("/")[-1]
        trigger = None
        for t in self.resources:
            if t["trigger_name"] in trigger_name:
                trigger = t
                break
        if not trigger:
            raise ValueError("No trigger found")
        response = trigger["func"](request)
        if not response:
            response = Response("success")
        return response

    def __add__(self, other):
        self.resources.extend(other.resources)
        return self

    def _deploy(self, sourceUrl=None, entrypoint=None):
        if not self.resources:
            return
        if self.config.eventarc and self.config.eventarc.get("serviceAccount"):
            service_account = self.config.eventarc.get("serviceAccount")
        elif self.config.cloudrun and self.config.cloudrun.get("service-account"):
            service_account = self.config.cloudrun.get("service-account")
        else:
            raise ValueError(
                "Service account not found for cloudrun or eventarc. You can set `serviceAccount` field in config.json under `eventarc`"
            )

        self.service_accounts = [service_account]

        log.info("deploying eventarc triggers......")
        for trigger in self.resources:
            topic = {}
            if trigger.get("topic"):
                topic = {
                    "transport": {
                        "pubsub": {
                            "topic": f"projects/{get_default_project()}/topics/{trigger.get('topic')}"
                        }
                    }
                }
            req_body = {
                "name": f"projects/{get_default_project()}/locations/{trigger['region']}/triggers/{trigger['trigger_name']}",
                "eventFilters": trigger["event_filters"],
                "serviceAccount": service_account,
                "destination": {
                    "cloudRun": {
                        "service": self.name,
                        "region": get_default_location(),
                        "path": f"/x-goblet-eventarc-triggers/{trigger['trigger_name']}",
                    }
                },
                "labels": self.config.labels,
                "eventDataContentType": trigger["event_data_content_type"],
                **topic,
            }
            create_eventarc_trigger(
                self.versioned_clients.eventarc,
                trigger["trigger_name"],
                trigger["region"],
                req_body,
            )

    def _sync(self, dryrun=False):
        """Only supports 1 region per sync"""
        triggers = self.versioned_clients.eventarc.execute(
            "list", parent_key="parent"
        ).get("triggers", [])
        filtered_triggers = list(
            filter(
                lambda trigger: f"/triggers/{self.name}-" in trigger["name"], triggers
            )
        )
        for filtered_trigger in filtered_triggers:
            filtered_name = filtered_trigger["name"].split("/")[-1]
            found = False
            for resource_trigger in self.resources:
                if resource_trigger["trigger_name"] == filtered_name:
                    found = True
                    break
            if not found:
                log.info(
                    f'Detected unused subscription in GCP {filtered_trigger["name"]}'
                )
                if not dryrun:
                    region = filtered_trigger["name"].split("/")[3]
                    destroy_eventarc_trigger(
                        self.versioned_clients.eventarc,
                        filtered_name,
                        region,
                    )

    def destroy(self):
        for trigger in self.resources:
            destroy_eventarc_trigger(
                self.versioned_clients.eventarc,
                trigger["trigger_name"],
                trigger["region"],
            )



import os
from goblet.client import VersionedClients
import goblet.globals as g
from goblet.common_cloud_actions import check_or_enable_service


class Infrastructure:
    """Base Infrastructure Class"""

    resource_type = ""
    can_sync = False
    supports_local = False
    required_apis = []
    permissions = []

    def __init__(
        self,
        name,
        backend=None,
        resources=None,
    ):
        self.name = name
        self.backend = backend
        self.resources = resources or {}
        self.config = g.config
        self.versioned_clients = VersionedClients()

    def register(self, name, kwargs):
        raise NotImplementedError("register")

    def deploy(self):
        if (
            not self.supports_local and os.getenv("X_GOBLET_LOCAL", False)
        ) and not os.getenv("G_HTTP_TEST") == "REPLAY":
            pass
        else:
            self._deploy()

    def _deploy(self):
        raise NotImplementedError("deploy")

    def destroy(self):
        raise NotImplementedError("destroy")

    def sync(self, dryrun=False):
        if self.can_sync:
            self._sync(dryrun)

    def _sync(self, dryrun=False):
        pass

    def get_config(self):
        return None

    def _check_or_enable_service(self, enable=False):
        if not self.resources:
            return
        return check_or_enable_service(self.required_apis, enable)

    def get_permissions(self):
        if len(self.resources) > 0:
            return self.permissions
        return []



import logging
import math
import os

from goblet.client import (
    get_default_project_number,
)

from goblet_gcp_client.client import get_default_location
from goblet.common_cloud_actions import deploy_cloudrun, get_artifact_image_name
from goblet.config import GConfig

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class RevisionSpec:
    def __init__(self, config={}, versioned_clients=None, name="goblet"):
        self.versioned_clients = versioned_clients
        if not isinstance(config, GConfig):
            config = GConfig(config=config)
        self.config = config
        self.cloudrun_configs = config.cloudrun or {}
        self.cloudrun_revision = config.cloudrun_revision or {}
        self.cloudrun_container = config.cloudrun_container or {}
        self.cloudrun_container["command"] = self.cloudrun_container.get("command") or [
            "functions-framework",
            "--target=goblet_entrypoint",
        ]
        self.req_body = {}
        self.artifactToDeploy = ""
        self.name = name

    def getServiceConfig(self):
        client = self.versioned_clients.run
        serviceConfig = client.execute(
            "get",
            parent_key="name",
            parent_schema=f"projects/{get_default_project_number()}/locations/{get_default_location()}/services/{self.name}",
            params={},
        )
        return serviceConfig

    # splits traffic proportionaly from already deployed traffic
    def modifyTraffic(self, serviceConfig={}):
        trafficSpec = self.cloudrun_configs.get("traffic")
        trafficList = []

        # proportion of total traffic specified
        trafficQuotient = (100 - trafficSpec) / 100
        # using the max for additional modifications
        maxTrafficVal = 0
        maxTrafficLoc = 0
        maxTraffic = {}
        # keep track of the total traffic
        trafficSum = 0

        for traffics in serviceConfig["trafficStatuses"]:
            newPercent = math.ceil(traffics["percent"] * trafficQuotient)

            if traffics["type"] == "TRAFFIC_TARGET_ALLOCATION_TYPE_LATEST":
                newTraffic = {
                    "type": "TRAFFIC_TARGET_ALLOCATION_TYPE_REVISION",
                    "revision": serviceConfig["latestReadyRevision"].rpartition("/")[
                        -1
                    ],
                    "percent": newPercent,
                }

            else:
                newTraffic = {
                    "type": "TRAFFIC_TARGET_ALLOCATION_TYPE_REVISION",
                    "revision": traffics["revision"],
                    "percent": newPercent,
                }

            trafficList.append(newTraffic)
            if traffics["percent"] > maxTrafficVal:
                maxTrafficLoc = len(trafficList) - 1
                maxTraffic = newTraffic
            trafficSum += newPercent

        latestRevisionTraffic = {
            "type": "TRAFFIC_TARGET_ALLOCATION_TYPE_LATEST",
            "percent": trafficSpec,
        }
        trafficList.append(latestRevisionTraffic)

        if trafficSpec > maxTrafficVal:
            maxTrafficLoc = len(trafficList) - 1
            maxTraffic = latestRevisionTraffic
            trafficSum += trafficSpec

        if trafficSum > 100:
            sub_from_max = trafficSum - 100
            maxTraffic["percent"] -= sub_from_max
            trafficList[maxTrafficLoc] = maxTraffic

        self.req_body["traffic"] = trafficList

    def deployRevision(self):
        client = self.versioned_clients.run
        region = get_default_location()
        project = get_default_project_number()

        self.artifactToDeploy = get_artifact_image_name(
            self.versioned_clients.cloudbuild, self.name, config=self.config
        )

        self.req_body = {
            "template": {
                **self.cloudrun_revision,
            },
            "labels": {**self.config.labels},
            **self.cloudrun_configs,
        }
        self.req_body["template"]["containers"] = [{**self.cloudrun_container}]
        self.req_body["template"]["containers"][0]["image"] = self.artifactToDeploy
        self.req_body["template"]["containers"].extend(
            self.config.cloudrun_container_extra or []
        )

        # check for traffic config
        if self.cloudrun_configs.get("traffic"):
            # check all services for the name of the service
            resp = client.execute(
                "list",
                parent_key="parent",
                parent_schema=f"projects/{project}/locations/{region}",
                params={},
            )

            for service in resp.get("services", []):
                if service["name"].rpartition("/")[-1] == self.name:
                    serviceConfig = self.getServiceConfig()
                    self.modifyTraffic(serviceConfig)

        deploy_cloudrun(client, self.req_body, self.name)



import logging
import os

from goblet.handlers.handler import Handler
from goblet_gcp_client.client import get_default_project, get_default_location
from goblet.common_cloud_actions import (
    get_function_runtime,
    create_cloudfunctionv2,
    create_cloudfunctionv1,
    destroy_cloudfunction,
)
from goblet.utils import nested_update

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))

STORAGE_EVENT_TYPES = {
    "v1": ["finalize", "delete", "archive", "metadataUpdate"],
    "v2": ["finalized", "deleted", "archived", "metadataUpdated"],
}


class Storage(Handler):
    """Storage trigger
    https://cloud.google.com/functions/docs/calling/storage
    """

    resource_type = "storage"
    valid_backends = ["cloudfunction", "cloudfunctionv2"]
    required_apis = ["cloudfunctions"]

    def __init__(self, name, backend, versioned_clients=None, resources=None):
        super(Storage, self).__init__(
            name,
            versioned_clients=versioned_clients,
            resources=resources,
            backend=backend,
        )
        self.resources = resources or []
        self.cloudfunction = f"projects/{get_default_project()}/locations/{get_default_location()}/functions/{name}"

    def validate_event_type(self, event_type):
        gcf_version = self.versioned_clients.cloudfunctions.version[:2]
        if event_type not in STORAGE_EVENT_TYPES[gcf_version]:
            raise ValueError(
                f"{event_type} not in {STORAGE_EVENT_TYPES[gcf_version]}. See https://cloud.google.com/functions/docs"
                f"/calling/storage for more information. "
            )

    def register(self, name, func, kwargs):
        bucket_name = kwargs["bucket"]
        event_type = kwargs["event_type"]
        self.validate_event_type(event_type)
        self.resources.append(
            {
                "bucket": bucket_name,
                "event_type": event_type,
                "name": name,
                "func": func,
            }
        )

    def __call__(self, event, context):
        event_type = context.event_type.split(".")[-1]
        bucket_name = event["bucket"]

        matched_buckets = [
            b
            for b in self.resources
            if b["bucket"] == bucket_name and b["event_type"][:6] == event_type[:6]
        ]
        if not matched_buckets:
            raise ValueError("No functions found")
        for b in matched_buckets:
            b["func"](event)

        return

    def __add__(self, other):
        self.resources.extend(other.resources)
        return self

    def _deploy(self, source=None, entrypoint=None):
        client = self.versioned_clients.cloudfunctions
        if not self.resources or not source:
            return

        log.info("deploying storage functions......")
        for bucket in self.resources:
            function_name = f"{self.cloudfunction}-storage-{bucket['name']}-{bucket['event_type']}".replace(
                ".", "-"
            )
            if self.versioned_clients.cloudfunctions.version == "v1":
                user_configs = self.config.cloudfunction or {}
                req_body = {
                    "name": function_name,
                    "description": self.config.description or "created by goblet",
                    "entryPoint": entrypoint,
                    "sourceUploadUrl": source["uploadUrl"],
                    "eventTrigger": {
                        "eventType": f"google.storage.object.{bucket['event_type']}",
                        "resource": f"projects/{get_default_project()}/buckets/{bucket['bucket']}",
                    },
                    "runtime": get_function_runtime(client, self.config),
                    "labels": self.config.labels,
                    **user_configs,
                }
                create_cloudfunctionv1(
                    self.versioned_clients.cloudfunctions, {"body": req_body}
                )
            elif self.versioned_clients.cloudfunctions.version.startswith("v2"):
                user_configs = self.config.cloudfunction_v2 or {}
                params = {
                    "body": {
                        "name": function_name,
                        "environment": "GEN_2",
                        "description": self.config.description or "created by goblet",
                        "buildConfig": {
                            "runtime": get_function_runtime(client, self.config),
                            "entryPoint": entrypoint,
                            "source": {"storageSource": source["storageSource"]},
                        },
                        "eventTrigger": {
                            "eventType": f"google.cloud.storage.object.v1.{bucket['event_type']}",
                            "eventFilters": [
                                {
                                    "attribute": "bucket",
                                    "value": bucket["bucket"],
                                }
                            ],
                        },
                        "labels": self.config.labels,
                    },
                    "functionId": function_name.split("/")[-1],
                }
                params["body"] = nested_update(params["body"], user_configs)
                create_cloudfunctionv2(self.versioned_clients.cloudfunctions, params)
            else:
                raise

    def destroy(self):
        for bucket in self.resources:
            destroy_cloudfunction(
                self.versioned_clients.cloudfunctions,
                f"{self.name}-storage-{bucket['name']}-{bucket['event_type']}".replace(
                    ".", "-"
                ),
            )



import logging
import os

from goblet.client import VersionedClients
from goblet_gcp_client.client import get_default_project, get_default_location
from goblet.common_cloud_actions import check_or_enable_service
import goblet.globals as g

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class Handler:
    """Base Handler class"""

    valid_backends = []
    resources = None
    resource_type = ""
    can_sync = False
    supports_local = False
    required_apis = []
    permissions = []
    service_accounts = []

    def __init__(
        self,
        name,
        backend,
        versioned_clients: VersionedClients = None,
        resources=None,
    ):
        self.config = g.config
        self.name = name
        self.backend = backend
        self.resources = resources or {}
        self.versioned_clients = versioned_clients or VersionedClients()
        self.cloudfunction = f"projects/{get_default_project()}/locations/{get_default_location()}/functions/{name}"

    def register(self, name, func, kwargs):
        raise NotImplementedError("register")

    def deploy(self, source=None, entrypoint=None):
        if self.resources and self.backend.resource_type not in self.valid_backends:
            log.info(
                f"skipping... {self.backend.resource_type} not supported for {self.resource_type}"
            )
            return
        if not self.resources:
            return
        if (
            not self.supports_local and os.getenv("X_GOBLET_LOCAL", False)
        ) and not os.getenv("G_HTTP_TEST") == "REPLAY":
            return
        self._deploy(source, entrypoint)
        try:
            self.set_invoker_permissions()
        except Exception:
            log.warning(f"Error setting invoker permissions for {self.resource_type}")

    def _deploy(self, source=None, entrypoint=None):
        raise NotImplementedError("deploy")

    def destroy(self):
        raise NotImplementedError("destroy")

    def sync(self, dryrun=False):
        if self.can_sync:
            log.info(f"syncing {self.resource_type}")
            self._sync(dryrun)

    def _sync(self, dryrun=False):
        pass

    def __call__(self, request, context=None):
        raise NotImplementedError("__call__")

    def __add__(self, other):
        if other.resources:
            self.resources.update(other.resources)
        return self

    def _check_or_enable_service(self, enable=False):
        if not self.resources:
            return
        return check_or_enable_service(self.required_apis, enable)

    def get_permissions(self):
        if len(self.resources) > 0:
            return self.permissions
        return []

    def set_invoker_permissions(self):
        if self.service_accounts:
            self.backend.add_invoker_binding(self.service_accounts)



import base64
import os
from goblet.common_cloud_actions import (
    create_pubsub_subscription,
    destroy_pubsub_subscription,
    get_function_runtime,
    create_cloudfunctionv2,
    create_cloudfunctionv1,
    destroy_cloudfunction,
)

import logging

from goblet.handlers.handler import Handler
from goblet_gcp_client.client import get_default_project
from goblet.utils import attributes_to_filter
from goblet.permissions import gcp_generic_resource_permissions, add_binding
from goblet.client import get_default_project_number
from googleapiclient.errors import HttpError

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class PubSub(Handler):
    """Pubsub topic trigger
    https://cloud.google.com/functions/docs/calling/pubsub
    """

    valid_backends = ["cloudfunction", "cloudfunctionv2", "cloudrun"]
    resource_type = "pubsub"
    can_sync = True
    supports_local = True
    required_apis = ["pubsub"]
    permissions = gcp_generic_resource_permissions("pubsub", "subscriptions")

    def register(self, name, func, kwargs):
        topic = kwargs["topic"]
        kwargs = kwargs.pop("kwargs")
        attributes = kwargs.get("attributes", {})
        config = kwargs.get("config", {})
        filter = kwargs.get("filter")
        if not filter and attributes:
            filter = attributes_to_filter(attributes)
        project = kwargs.get("project", get_default_project())
        deploy_type = "trigger"
        if (
            kwargs.get("use_subscription")
            or project != get_default_project()
            or self.backend.resource_type == "cloudrun"
        ):
            deploy_type = "subscription"

        if self.resources.get(topic):
            self.resources[topic][deploy_type][name] = {
                "func": func,
                "attributes": attributes,
                "project": project,
                "filter": filter,
                "force_update": kwargs.get("force_update", False),
            }
        else:
            self.resources[topic] = {"trigger": {}, "subscription": {}}
            self.resources[topic][deploy_type] = {
                name: {
                    "func": func,
                    "attributes": attributes,
                    "project": project,
                    "filter": filter,
                    "config": config,
                    "force_update": kwargs.get("force_update", False),
                }
            }

    def __call__(self, event, context):
        # Trigger
        if context:
            try:
                topic_name = context.resource.split("/")[-1]
            except AttributeError:
                topic_name = context.resource["name"].split("/")[-1]
            data = base64.b64decode(event["data"]).decode("utf-8")
            attributes = event.get("attributes") or {}
        # Subscription
        else:
            subscription = event.json["subscription"].split("/")[-1]
            topic_name = subscription.replace(self.name + "-", "")
            data = base64.b64decode(event.json["message"]["data"]).decode("utf-8")
            attributes = event.json["message"].get("attributes") or {}

        topic = self.resources.get(topic_name)
        if not topic:
            raise ValueError(f"Topic {topic_name} not found")

        # check attributes
        response = None
        for _, info in topic["trigger"].items():
            if info["attributes"].items() <= attributes.items():
                response = info["func"](data)
        for _, info in topic["subscription"].items():
            if info["attributes"].items() <= attributes.items():
                response = info["func"](data)
        return response or "success"

    def _deploy(self, source=None, entrypoint=None):
        if not self.resources:
            return
        for topic_name in self.resources:
            # Deploy triggers
            for _, topic_info in self.resources[topic_name]["trigger"].items():
                self._deploy_trigger(
                    source=source, entrypoint=entrypoint, topic_name=topic_name
                )
            # Deploy subscriptions
            for _, topic_info in self.resources[topic_name]["subscription"].items():
                self._deploy_subscription(topic_name=topic_name, topic=topic_info)

    def _deploy_subscription(self, topic_name, topic):
        sub_name = f"{self.name}-{topic_name}"
        log.info(f"deploying pubsub subscription {sub_name}......")

        if os.environ.get("X_GOBLET_LOCAL"):
            push_url = os.environ.get("GOBLET_LOCAL_URL", "http://localhost:8080")
        else:
            push_url = self.backend.http_endpoint

        if self.config.pubsub and self.config.pubsub.get("serviceAccountEmail"):
            service_account = self.config.pubsub.get("serviceAccountEmail")
        elif (
            self.backend.resource_type == "cloudrun"
            and self.config.cloudrun_revision
            and self.config.cloudrun_revision.get("serviceAccount")
        ):
            service_account = self.config.cloudrun_revision.get("serviceAccount")
        elif self.backend.resource_type.startswith("cloudfunction") and (
            self.config.cloudfunction or self.config.cloudfunction_v2
        ):
            service_account = self.config.pubsub.get("serviceAccountEmail")
        else:
            raise ValueError(
                "Service account not found in cloudrun or cloudfunction. You can set `serviceAccountEmail` field in config.json under `pubsub`"
            )

        self.service_accounts = [service_account]
        req_body = {
            "topic": f"projects/{topic['project']}/topics/{topic_name}",
            "filter": topic["filter"] or "",
            "pushConfig": {}
            if topic["config"].get("enableExactlyOnceDelivery", None)
            else {
                "pushEndpoint": push_url,
                "oidcToken": {
                    "serviceAccountEmail": service_account,
                    "audience": push_url,
                },
            },
            "labels": self.config.labels,
            **topic["config"],
        }
        deadLetterPolicy = topic["config"].get("deadLetterPolicy", {})
        if deadLetterPolicy != {}:
            req_body["deadLetterPolicy"] = deadLetterPolicy
        create_pubsub_subscription(
            client=self.versioned_clients.pubsub,
            sub_name=sub_name,
            req_body=req_body,
            force_update=topic["force_update"],
        )
        # Add IAM roles to use dead-letter topics
        if deadLetterPolicy != {}:
            try:
                default_pubsub_service_account = f"serviceAccount:service-{get_default_project_number()}@gcp-sa-pubsub.iam.gserviceaccount.com"
                add_binding(
                    self.versioned_clients.pubsub,
                    "projects/{project_id}/subscriptions/" + sub_name,
                    "roles/pubsub.subscriber",
                    default_pubsub_service_account,
                )
            except HttpError as e:
                if e.resp.status == 403:
                    log.info(
                        f"User is not authorized to add IAM role 'roles/pubsub.subscriber' to subscription '{sub_name}' you need to handle this manually."
                    )

    def _deploy_trigger(self, topic_name, source=None, entrypoint=None):
        function_name = f"{self.cloudfunction}-topic-{topic_name}"
        log.info(f"deploying topic function {function_name}......")
        if self.versioned_clients.cloudfunctions.version == "v1":
            user_configs = self.config.cloudfunction or {}
            req_body = {
                "name": function_name,
                "description": self.config.description or "created by goblet",
                "entryPoint": entrypoint,
                "sourceUploadUrl": source["uploadUrl"],
                "eventTrigger": {
                    "eventType": "providers/cloud.pubsub/eventTypes/topic.publish",
                    "resource": f"projects/{get_default_project()}/topics/{topic_name}",
                },
                "runtime": get_function_runtime(
                    self.versioned_clients.cloudfunctions, self.config
                ),
                "labels": self.config.labels,
                **user_configs,
            }
            create_cloudfunctionv1(
                self.versioned_clients.cloudfunctions, {"body": req_body}
            )
        elif self.versioned_clients.cloudfunctions.version.startswith("v2"):
            user_configs = self.config.cloudfunction_v2 or {}
            params = {
                "body": {
                    "name": function_name,
                    "environment": "GEN_2",
                    "description": self.config.description or "created by goblet",
                    "buildConfig": {
                        "runtime": get_function_runtime(
                            self.versioned_clients.cloudfunctions, self.config
                        ),
                        "entryPoint": entrypoint,
                        "source": {"storageSource": source["storageSource"]},
                    },
                    "eventTrigger": {
                        "eventType": "google.cloud.pubsub.topic.v1.messagePublished",
                        "pubsubTopic": f"projects/{get_default_project()}/topics/{topic_name}",
                    },
                    "labels": self.config.labels,
                    **user_configs,
                },
                "functionId": function_name.split("/")[-1],
            }
            create_cloudfunctionv2(self.versioned_clients.cloudfunctions, params)
        else:
            raise

    def _sync(self, dryrun=False):
        subscriptions = self.versioned_clients.pubsub.execute(
            "list", parent_key="project"
        ).get("subscriptions", [])
        filtered_subscriptions = list(
            filter(
                lambda sub: f"subscriptions/{self.name}-" in sub["name"], subscriptions
            )
        )

        for filtered_sub in filtered_subscriptions:
            split_name = filtered_sub["name"].split("/")[-1].split("-")
            filtered_name = split_name[1]
            if not self.resources.get(filtered_name):
                log.info(f'Detected unused subscription in GCP {filtered_sub["name"]}')
                if not dryrun:
                    destroy_pubsub_subscription(
                        self.versioned_clients.pubsub, f"{self.name}-{filtered_name}"
                    )

    def is_http(self):
        """
        Http cloudfunction is needed for cloudfunction subscription
        """
        for _, topic_info in self.resources.items():
            if topic_info.get("subscription"):
                return True
        return False

    def destroy(self):
        if not self.resources:
            return
        for topic_name in self.resources:
            # Destroy triggers
            for _, topic_info in self.resources[topic_name]["trigger"].items():
                destroy_cloudfunction(
                    self.versioned_clients.cloudfunctions,
                    f"{self.name}-topic-{topic_name}",
                )
            # Destroy subscriptions
            for _, topic_info in self.resources[topic_name]["subscription"].items():
                destroy_pubsub_subscription(
                    self.versioned_clients.pubsub, f"{self.name}-{topic_name}"
                )



from goblet.handlers.handler import Handler

import logging
import os

log = logging.getLogger("goblet.deployer")
log.setLevel(logging.getLevelName(os.getenv("GOBLET_LOG_LEVEL", "INFO")))


class CloudTaskTarget(Handler):
    """CloudTask Target"""

    resource_type = "cloudtask_http_target"
    valid_backends = ["cloudfunction", "cloudfunctionv2", "cloudrun"]

    def register(self, name, func, kwargs):
        """
        @app.cloudtasktarget(name="target")
        def my_handler(request):
            pass

        self.resources["target"] = {
            "func": <function my_handler at 0x...>
        }
        """
        if name in self.resources.keys():
            raise Exception(f"cloudtasktarget {name} already registered")

        self.resources[name] = {
            "func": func,
        }

    def __call__(self, request, context=None):
        target = request.headers.get("X-Goblet-CloudTask-Target", None)
        if not target:
            raise ValueError("No X-Goblet-CloudTask-Target header found")

        try:
            return self.resources[target]["func"](request)
        except KeyError:
            log.info(f"{target} not found")
        except Exception as e:
            raise e

    def _deploy(self, source=None, entrypoint=None, config={}):
        self.service_accounts = self.config.cloudtask.get("serviceAccount", [])
        return

    def destroy(self):
        return
